/* all this stuff is new for the 6502 version */


#include <stdio.h>
#include <string.h>
#include "cc65.h"

/* flags for enabling/disabling parts of the optimizer.  mostly so
   jrd can figure out the most effective combination... */
/*
#define OPT12 
*/
	 
int correct;	/* set true if the next ..ax should turned into ..a */

int matchstr(char * line,char * str);
int matchjsr(char * line,char * template);
char *nextline(char * line);
char *prepend(char * str1,char * str2);
char * optpair(char * ptr,char * rtn1,char * rtn2,char * rtn3);
char *opt_trip(char * ptr);
char *opt11(char * ptr);
#ifdef OPT12
char * opt12(char * ptr);
#endif
char * opt(char * ptr);
void peephole(char * ptr);

/*io.c*/
void sout(char *s);
void cout(char ch);



char * findstr(char * line, char *str)
{
char * h,*h1 =str;
  while ( *str )
  {
    while ( *line && *line != '\n' && *line != *str)
      ++line;
      if (!*line || *line == '\n') return 0;
		
      h = line;
      while ( *line && *line != '\n' && *line == *str && *str) 
      {  ++str; ++line; }
      if ( !*str) return h;
        str = h1;
	++line;
  }
}		    
			

/* return t if line looks like " jsr <template>" */
int matchjsr(char * line,char * template)
{
  if (!matchstr(line, "\tjsr\t"))
    return (0);
  return (matchstr(line + 5, template));
}

/* handed a pointer to beginning of line, return pointer to beginning
   of next line, or NIL, if there isn't one */
   
char *nextline(char * line)
{
  for (; *line; ++line)
    if (*line == '\n')
      if (line[1] == '\0')
	return (0);
      else
	return (line + 1);
  return (0);
}


/* return ptr to str1 prepended to str2 */
char *prepend(char * str1,char * str2)
{
char * r;

  r = str2 = str2 - strlen(str1);
  for (; *str1;)
    {
      *str2 = *str1;
      ++str2;
      ++str1;
    }
  return (r);
}

/* return ptr to terminator of line */
/* ...

char * findeol(char * l)
{
  while (*l != '\n')
    ++l;
  return (l);
}

...*/

/* handy helper fun used by a lot of optimizers below.  if find
   jsr matches of the two passed values, substitute the third value.
   NB!! returns 0 if didn't change anything! */
char * optpair(char * ptr,char * rtn1,char * rtn2,char * rtn3)
{
char * l;

  if ((l = nextline(ptr)) == 0)
    return (ptr);
  if (matchjsr(ptr, rtn1) && matchjsr(l, rtn2))
    {
/*
	printmsg("match '%s'", rtn1);
	printmsg(", '%s'", rtn2);
	printmsg("-> '%s'\n", rtn3);
*/
      return (prepend("\tjsr\t", prepend(rtn3, strchr(l, '\n'))));
    }
  return ( (char *) NULL);
}


/* actual optimizer routines.
   The basic idea here is that we get a pointer, find some number of
   next-line pointers, and look for obvious patterns that can be
   replaced by simpler ones.  when we find one, prepend the simpler
   pattern, and return the pointer to new beginning of buf. */

extern char * triples[];

char *opt_trip(char * ptr)
{
char * new_ptr;
#ifdef smallc
int * triple;
#else
char ** triple;
#endif

  for (triple = triples; triple[0]; triple += 3)
    {

	printf("optimizing triple '%s' '%s' '%s'\n",
		triple[0], triple[1], triple[2]);

      if ( (new_ptr = optpair(ptr, triple[0], triple[1], triple[2])) )
	{
	  ptr = new_ptr;
	}
    }
  return (ptr);
}

/* optimize ldax #0 followed by pushax into push0 */
/* and ldax #(0<n<256) by lda #n; pusha           */
char *opt11(char * ptr)
{
char * l;
int i;

  if ((l = nextline(ptr)) == 0 ) return (ptr);

  if ( !matchstr(ptr,"\tldax") ) return (ptr);
  
  correct = 0;
  
  if ( matchjsr(l,"pushax") && ptr[7] == '0')
    return (prepend("\tjsr\tpush0", strchr(l, '\n')));

  if ( (i = atoi(ptr+7)) > 255 || !i) return (ptr);
  
  if (! matchstr(l,"\tjsr\t") ) return (ptr);

  ptr[4]=' ';
  correct = 1;
  return (ptr);
}

#ifdef OPT12
/* optimize "axzerop" followed by lbeq, to just reverse the sense
   of the test
*/

char * opt12(char * ptr)
{
char * l;
int i;

  if ((l = nextline(ptr)) == 0)
    return (ptr);
  if (matchstr(l, "\tlbeq") && matchjsr(ptr, "axzerop"))
    {
      correct = 0;
      l[3] = 'n'; /* make it lbne */
      l[4] = 'e';
      return (l);		/* return the lbne */
    }
  return (ptr);
}

#endif
//
// optimize constant asl
//
static char LDY[] = "\tldy\t#$10";
static char inthex[] = "0123456789ABCDEF";
char * opt13(char * ptr)
{
char * l,*l1,*l2;
int count;

  if ( (l = nextline(ptr)) == 0)
    return ptr;
  if ( (l1 = nextline(l)) == 0)
    return ptr;

  if ( matchstr(l,"\tldax\t#") && matchjsr(l1,"asltos") )
    {
      correct = 0;
      count = atoi(strchr(l,'#')+1);
      if ( count == 1)
        if ( matchjsr(ptr,"pushax"))
        {
          return (prepend("\tjsr aslax",strchr(l1,'\n')));
        }
        else
	      { 
	        correct= 1;
          return (prepend("\tasl\tA",strchr(l1,'\n')));
	      }
      else
        if ( count < 16 )
        {
	        LDY[7] = inthex[count]; LDY[8] = ' ';
          if (matchjsr(ptr,"pushax"))
            return (prepend(LDY,prepend("\n\tjsr aslaxy",strchr(l1,'\n'))));
	        else
            return (prepend(LDY,prepend("\n\tjsr aslay",strchr(l1,'\n'))));
        }
     }
    
  return ptr;
}
//
// optimize constant asr 
//
char * opt14(char * ptr)
{
char * l,*l1,*l2;
char help[4];
int count;

  if ( (l = nextline(ptr)) == 0)
    return ptr;
  if ( (l1 = nextline(l)) == 0)
    return ptr;

  if ( matchstr(l,"\tldax\t#") && matchjsr(l1,"asrtos") )
    {
      correct = 0;
      count = atoi(strchr(l,'#')+1);
      if ( count == 1)
        if (matchjsr(ptr,"pushax"))
        {
          return (prepend("\tjsr asrax",strchr(l1,'\n')));
        }
	      else
	      {
          correct = 1;
          return (prepend("\tcmp\t#$80\n\tror\tA",strchr(l1,'\n')));
        }
      else
        if ( count < 16 )
        {
          LDY[7] = inthex[count];LDY[8] = ' ';
          if (matchjsr(ptr,"pushax"))
            return (prepend(LDY,prepend("\n\tjsr asraxy",strchr(l1,'\n'))));
	        else
            return (prepend(LDY,prepend("\n\tjsr asray",strchr(l1,'\n'))));
        }
     }
  return ptr;
}
//
// turn ldx #0 ; ...; lda xxx into ;...; lda xxx 
//
char * opt15(char * ptr)
{
char * l,*l1,*l2;
int flag = 0;

  if ( (l = nextline(ptr)) == 0) return (ptr);

  if ( ! matchstr(ptr,"\tldx\t#0") ) return (ptr);

  if ( (l1 = nextline(l)) == 0) return ptr;

  if ( matchstr(l,"\tldy\t#") )
    if ( (l1 = nextline(l1)) == 0)
      return ptr;

  if ( matchstr(l1,"\tjsr\t") && !matchstr(l1,"\tjsr\tincsp") )
  {
    correct = 1;
    return ( l );
  }

  return ptr;
}
//
// change ldax #xx; sta => lda #xx; sta ..
//
char * opt16(char * ptr)
{
char *l,*l1;

  if ( (l = nextline(ptr)) == 0) return ptr;

  if ( (l1 = nextline(l)) == 0) return ptr;

  if ( matchstr(ptr,"\tldax\t#") && matchstr(l,"\tsta") && !matchstr(l1,"\tstx") ) 
  {	
    correct = 1;
    ptr[4]=' ';     
  }
  return (ptr);
}
//
// change byte incax1 or decax1 to ina/dea
//
char * opt17(char * ptr)
{
char *l,*l1,*l2;
int flag = 0;

  if ( !matchstr(ptr,"\tldx\t#0") ) return (ptr);

  if ( (l = nextline(ptr)) == 0) return ptr;

  if ( matchstr(l,"\tldy\t#") )
  {
     if ( (l2 = nextline(l )) == 0) return (ptr);
     if ( (l1 = nextline(l2)) == 0) return (ptr);
     flag = 1;
  }else
   if ( (l1 = nextline(l)) == 0) return ptr;

 if ( matchjsr(l1,"incax1") || matchjsr(l1,"decax1") )
  {
     correct = 1;
     l1[1] =l1[5];
     l1[2] =l1[6];
     l1[3] = 'a';
     l1[4] = ';';
     if (flag && (l1=nextline(l1)) != 0)
       *l1 = ';';
     return l;
   }
  return ptr;
}
//
// optimize byte stores
//
char * opt18(char * ptr)
{
char *l,*l1,*l2;

  if ( !matchstr(ptr,"\tldx\t#0") ) return (ptr);

  if ( (l = nextline(ptr)) == 0) return ptr; /* lda */ 

  if ( (l1 = nextline(l)) == 0)  return ptr; /* sta */
 
  if ( matchstr(l,"\tldy\t#") )
     if ( (l1 = nextline(l1)) == 0) return (ptr); /* lda (sp),y */

  if ( matchstr(l1,"\tldy\t#") )
    if ( (l1 = nextline(l1)) == 0) return (ptr); /* sta (sp),y */

  if ( (l2 = nextline(l1)) == 0) return (ptr);
    
  if ( matchstr(l1,"\tsta") && !matchstr(l2,"\tstx") )
  {
    correct = 1;
    return (l);
  }

  return ptr;
} 
//
// change lbra/lbeq to lbne
//
char * opt19(char *ptr)
{
char *l,*l1;
int num;

  if ( (l=nextline(ptr))==0) return (ptr);

  if ( (l1=nextline(l))==0)  return (ptr);

  if (matchstr(ptr,"\tlbeq\t") && matchstr(l,"\tlbra\t") )
  {
    correct = 0;
    num = atoi(strchr(ptr,'L')+1);
    if ( (*l1 =='L') && (atoi(l1+1)==num) )
    {
      l[3] = 'n'; l[4] = 'e';
      return (l);
    }
  }
  return ptr;
}
//
// change constant addition to indexax or incax1...7 or adc ..
//
char incax[]="\tjsr\tincax1";
char ADC[]="\tclc\n\tadc\t#$00";
char indexax[]="\tldy\t#$00\n\tjsr\tindexax\n";
char * opt20(char * ptr)
{
char * l,*l1,*l2;
int count;

  if ( (l = nextline(ptr)) == 0) return ptr;

  if ( (l1 = nextline(l)) == 0)  return ptr;

  if ( matchstr(l,"\tldax\t#") && matchjsr(l1,"addtos") )
  {
    correct = 0;
    
    count = atoi(strchr(l,'#')+1);

    if ( ! count )
      if ( (l = nextline(l1)) != NULL)
        return l;
      else
        return ptr;

    if (count > 255) return (ptr);

    if ( ! count )
      if ( (l1 = nextline(l1)) )
        return l1;
      else
        return ptr;

    if (matchjsr(ptr,"pushax") )
    {
      if ( count <= 8 )
      {
        incax[10]=inthex[count & 15];
        return (prepend(incax,strchr(l1,'\n')));
      }
      else
      {
        indexax[7] = inthex[count>>4];indexax[8] =inthex[count & 15]; 
        return (prepend(indexax,strchr(l1,'\n')));
      }
    }
    else
    {
      correct = 1;
      if ( count > 1 )
      {
        ADC[7+5] = inthex[count>>4];ADC[8+5] =inthex[count & 15]; 
        return (prepend(ADC,strchr(l1,'\n')));
      }
      else
        return (prepend("\tina",strchr(l1,'\n')));
     }
  }
  return ptr;
}

char decax[]="\tjsr\tdecax1";
char SBC[]="\tsec\n\tsbc\t#$00";
char subaxy[]="\tldy\t#$00\n\tjsr\tsubaxy";
char * opt21(char * ptr)
{
char * l,*l1,*l2;
int count;

  if ( (l = nextline(ptr)) == 0) return ptr;

  if ( (l1 = nextline(l)) == 0)  return ptr;

  if ( matchstr(l,"\tldax\t#") && matchjsr(l1,"subtos") )
    {
      correct = 0;

      count = atoi(strchr(l,'#')+1);

      if (count > 255) return (ptr);
      
      if ( ! count )
        if ( (l1 = nextline(l1)) )
          return l1;
        else
          return ptr;
      
      if (matchjsr(ptr,"pushax") )
      {
        if ( count <= 2 )
	      {
          decax[10]=inthex[count & 15];
          return (prepend(decax,strchr(l1,'\n')));
	      }
	      else
	      {
	        subaxy[7] = inthex[count>>4];subaxy[8] =inthex[count & 15]; 
          return (prepend(subaxy,strchr(l1,'\n')));
	      }
      }
      else
      {
	      correct = 1;
	      if ( count > 1 )
	      {
	        SBC[7+5] = inthex[count>>4];SBC[8+5] =inthex[count & 15]; 
	        return ( prepend(SBC,strchr(l1,'\n')));
	      }else
	        return (prepend("\tdea",strchr(l1,'\n')));
      }
    }
 return ptr;
}

/* change word (ax) into byte versions of runtime-calls */
/* these calls set X=0                                  */

char * DoCorrect(char *ptr)
{
char *l,*l1;

  if ( (l=nextline(ptr))==0 ) return (ptr);

  if ( matchstr(ptr,"\tlda") ||
       matchstr(ptr,"\tldy") ||
       matchstr(ptr,"\tpha") )
  {
    return ptr;
  }

  if ( matchstr(ptr,"\tlb") )   // lbra || lbne || lbeq
  {
    correct = 0;
    return (ptr);
  }

  if ( !matchstr(ptr,"\tjsr") ) return ptr;

  if ( matchjsr(ptr,"stax") || matchjsr(ptr,"ldax"))
  {
    ptr[8]='_';
    correct = 0;
    return (ptr);
  }
  
  if ( l1 = findstr(ptr,"ax")  )
  {
    while( l1[2] != '\n')
    {
       l1[1]=l1[2];
       ++l1;
    }
    l1[1]=' ';
    correct = 0;
  }else if ( l1 = findstr(ptr,"tos") )
  {
    l1[2] = 'a';
    correct = 0;
  }
  correct = 0;
  return ptr;
} 

/* the optimizer main loop */
char * opt(char * ptr)
{
char * lastptr;

  lastptr = 0;
  while (lastptr != ptr)
    {
    lastptr = ptr;

    /*ptr = opt_trip(ptr);*/

#ifdef OPT12
//    ptr = opt12(ptr);
#endif

    ptr = opt13(ptr);	/* asl #const */
    ptr = opt14(ptr);	/* asr #const */
    ptr = opt17(ptr); /* optimizes ++a or --a */
    ptr = opt19(ptr); /* optimize lbne/lbra => lbeq */ 
    ptr = opt20(ptr);	/* constant addition */
    ptr = opt21(ptr);	/* constant substraction */
    ptr = opt15(ptr);	/* ldx #0 ; lda xx ; jsr ...  => lda #xx + correct = 1*/
    ptr = opt18(ptr); /* optimize byte moves */
    ptr = opt16(ptr);	/* ldax #0 ; sta xx => lda # ;sta xx */
    ptr = opt11(ptr);   /* ldax #const => lda #const V const <256 */
                        /* and ldax #0;pushax => push0            */
  }
    if (correct)
      ptr = DoCorrect(ptr);	/* if one of the above produces byte, and */
                            /* next expr. needs word this is needed   */
  /* more later */
  
  return (ptr);
}

static char xxc;
void peephole(char * ptr)
{
  correct = 0;
  if (optimize)
    while ( (xxc = *ptr) )
      {
	    if (xxc == '\t') /* only check obvious lines */
	      ptr = opt(ptr);
	    while ((xxc = *ptr) && (xxc != '\n'))
	    {
	      cout(xxc);
	      ++ptr;
	    }
	    if (xxc)
	    {
	      cout(xxc);
	      ++ptr;
	    }
    }
  else
    sout(ptr);
}
