
/*
   This software is copyright 1989 by John Dunning.  See the file
   'COPYLEFT.JRD' for the full copyright notice.
*/

/* symbol table code */

#include "util.h"
#include "symtab.h"
#include "parse.h"
#include "global.h"

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

extern void barf(char * msg,int arg1,int arg2,int arg3);

#define	HASHMOD	512

/* #define CASE_INSENSITIVE */

SYM* symtab[HASHMOD];

void string_upcase(str)
char * str;
{
#ifdef CASE_INSENSITIVE
  char c;

  for( ; ( c = *str ) ; str++)
	if ((c >= 'a') && (c <= 'z'))
		*str = c - 0x20;
#endif
}
  
int hash(register char *str)
{
register  int accum;
register  char c;

  for (accum = 0 ; ( c = *str++ ); )
	{
#ifdef CASE_INSENSITIVE
	c = toupper(c);
#endif
	accum = ((accum << 1) & 0x3FFF) ^ c;
	}
  return(accum & (HASHMOD-1));
}

SYM * find_sym(name, new_ok)
char * name;
int new_ok;
{
  SYM * sym_p;
  int idx;

  idx = hash(name);
  for (sym_p = symtab[idx] ; sym_p ; sym_p = sym_p->next)
	if (string_equal(sym_p->name, name))
		return(sym_p);

  if (!new_ok)	return(0);

  sym_p = (SYM* )malloc((sizeof (SYM)) + strlen(name) + 1);
  sym_p->value = 0;
  sym_p->flags = 0;
  sym_p->nbr = -1;
  sym_p->line = line_nbr;
  strcpy(sym_p->name, name);
  string_upcase(sym_p->name);
  sym_p->next = symtab[idx];
  symtab[idx] = sym_p;
  return(sym_p);
}

SYM * assign_sym(name, value, no_redef)
char * name;
int value;
int no_redef;
{
  SYM * sym_p;

  sym_p = find_sym(name, 1);
  if (no_redef && (sym_p->flags & DEFINED))
	if (sym_p->value != value)
		barf("'%s' redefined: %X vs %X\n",(int) name,(int)sym_p->value, value);
  sym_p->value = value;
  sym_p->flags |= DEFINED;
  return(sym_p);
}

void set_label()
{
  if (p.label)
	{
	assign_sym(p.label, pc, 1);
	if (!p.opcode)
		{
		list_v = pc;
		list_v_p = 1;
		}
	    else
		{
		list_pc = pc;
		}
	}
}

void init_sym()
{
  int i;

  for (i = 0 ; i < HASHMOD ; i++)
	symtab[i] = 0;
}

void describe_symtab()
{
  int i, j;
  SYM * sym_p;

  for (i = 0 ; i < HASHMOD ; i++ )
	{
	for (j = 0, sym_p = symtab[i] ; sym_p ; j++)
		sym_p = sym_p->next;
	printf("Bucket %d contains %d syms\n", i, j);
	}
}

void map_syms(void (* fun)())
 
{
  SYM * sym_p;
  int i;

  for (i = 0 ; i < HASHMOD ; i++ )
	{
	for (sym_p = symtab[i] ; sym_p ; )
		{
		(*fun)(sym_p);
		sym_p = sym_p->next;
		}
	}
}
extern char * in_name[];
int find_udef()
{
  SYM * sym_p;
  int i;
  int udef = 0;
  int save_line = line_nbr;
  
  for (i = 0 ; i < HASHMOD ; i++)
    for (sym_p = symtab[i]; sym_p ; )
    {
      if ( !(sym_p->flags & DEFINED) && !(sym_p->flags & GLOBAL) )
      {
        line_nbr = sym_p->line;
//not yet        strcpy(CurrentFile,in_name[sym_p->nbr]);
        barf("'%s' undefined (first use)\n",(int)sym_p->name,0,0);
        ++udef;
      }
      sym_p = sym_p->next;
    }
  line_nbr = save_line;
  return udef;
}

#ifdef DEBUG
void dump_syms()
{
  char * name[32];
  int i, j;
  SYM * sym_p;
  FILE * f;

  sprintf(name, "pass%d.sym", pass);
  f = fopen(name, "w");
  for (i = 0 ; i < HASHMOD ; i++ )
	{
	for (j = 0, sym_p = symtab[i] ; sym_p ; j++)
		{
		fprintf(f, "%s:\t\t%X\n", sym_p->name, sym_p->value);
		sym_p = sym_p->next;
		}
	}
  fclose(f);
}
#endif

